/* sws.c: Simple (Silly?) Web Server */
/* This software is released under the GPL, see GPL.txt for more info */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <getopt.h>
#include <time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <assert.h>

#ifndef DEFAULT_PORT
#define DEFAULT_PORT 8642
#endif

#ifndef DEFAULT_LOGFILE
#define DEFAULT_LOGFILE "/dev/null"
#endif

#define DOCBUFSIZE 4
#define BUFFERSIZE (DOCBUFSIZE * 1024)

#define VERSION "SWS/0.1"

static char *docbuf;

static char *doc404 = "HTTP/1.1 404 Not Found\n"
    "Server: " VERSION "\n"
    "Connection: close\n"
    "Content-Type: text/html\n"
    "\n"
    "<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">\n"
    "<HTML><HEAD>\n"
    "<TITLE>404 Not Found</TITLE>\n"
    "</HEAD><BODY>\n"
    "<H1>Not Found</H1>\n"
    "The requested URL was not found on this server.\n"
    "<P><HR><ADDRESS>" VERSION "</ADDRESS>\n" "</BODY></HTML>\n";

static char *ok_header = "HTTP/1.1 200 OK\n"
    "Server: " VERSION "\n"
    "Last-Modified: Wed, 1 Jan 1999 00:00:00 GMT\n"
    "Connection: close\n" "Content-Type: text/html\n";

static char *ok_leadin =
    "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2 Final//EN\">\n"
    "<HTML>\n" "<HEAD>\n" "<TITLE>" VERSION " Test Page</TITLE>\n"
    "</HEAD>\n" "<BODY>\n";

static char *ok_leadout = "\n" "</BODY>\n" "</HTML>\n";

static int ok_header_len, ok_leadin_len, ok_leadout_len;

static volatile int sigpipe, sigterm;
static FILE *logfp;


static void logmsg(const char *msg)
{
    time_t t = time(NULL);
    struct tm *curr_time = localtime(&t);
    char loctime[80];

    strcpy(loctime, asctime(curr_time));
    loctime[strlen(loctime) - 1] = '\0';
    fprintf(logfp, "%s: pid %d: %s\n", loctime, getpid(), msg);
    fflush(logfp);
}


static void sigpipe_handler(int sig)
{
    sigpipe = 1;
}


static void sigterm_handler(int sig)
{
    char logbuf[100];
    snprintf(logbuf, 100, "Got signal %d, terminating", sig);
    logmsg(logbuf);
    sigterm = 1;
}


static int init_buffer(void)
{
    int i;

    docbuf = (char *) malloc(BUFFERSIZE);
    if (!docbuf) {
	return -1;
    }

    memset(docbuf, 'x', BUFFERSIZE);
    for (i = 64; i <= BUFFERSIZE; i += 64) {
	strncpy(docbuf + i - 5, "<BR>\n", 5);
    }

    return 0;
}


#define REQBUFSIZ 4096
#define MAXDOCNAMELEN 1024

static int get_request(int sock)
{
    char reqbuf[REQBUFSIZ], doc[MAXDOCNAMELEN];
    unsigned int offset = 0;
    int prot_major, prot_minor;

    while (offset < REQBUFSIZ) {
	int got = recv(sock, reqbuf + offset, REQBUFSIZ - offset, 0);
	if (got <= 0) {
	    if (got < 0) {
		perror("recv");
		return -1;
	    }
	    logmsg("Connection closed by client");
	    fprintf(stderr, "Connection closed by client\n");
	    return -1;
	}
	offset += got;
	if (offset > 1) {
	    if (reqbuf[offset - 1] == '\n' && reqbuf[offset - 2] == '\n') {
		break;
	    }
	}
	if (offset > 3) {
	    if (reqbuf[offset - 1] == '\n' && reqbuf[offset - 2] == '\r' &&
		reqbuf[offset - 3] == '\n' && reqbuf[offset - 4] == '\r') {
		break;
	    }
	}
    }

    if (offset >= REQBUFSIZ) {
	logmsg("Request buffer overflow, closing connection");
	fprintf(stderr, "Request buffer overflow, closing connection\n");
	return -1;
    }

    if (sscanf(reqbuf, "GET %s HTTP/%d.%d", doc, &prot_major, &prot_minor) == 3) {
	char logbuf[1024];
	int docsize;
	char unit;

	if (sscanf(doc, "/file%d%c", &docsize, &unit) == 2 &&
	    (unit == 'k' || unit == 'm')) {
	    if (docsize < 0) {
		docsize = 0;
	    } else if (unit == 'm') {
		docsize *= 1024;
	    }
	    if (docsize > 0) {
		snprintf(logbuf, 1024, "\"GET %s HTTP/%d.%d\" OK",
			 doc, prot_major, prot_minor);
	    } else {
		snprintf(logbuf, 1024, "\"GET %s HTTP/%d.%d\" Not found",
			 doc, prot_major, prot_minor);
	    }
	    logmsg(logbuf);
	    return docsize;
	} else {
	    snprintf(logbuf, 1024, "\"GET %s HTTP/%d.%d\" Not found",
		     doc, prot_major, prot_minor);
	    logmsg(logbuf);
	    return 0;
	}
    } else {
	logmsg("Got illegal request");
	fprintf(stderr, "Got illegal request\n");
	return -1;
    }
}


static int send_all(int sock, char *buffer, int count)
{
    do {
	int written = send(sock, buffer, count, 0);
	if (written < 0) {
	    if (sigpipe) {
		sigpipe = 0;
		signal(SIGPIPE, sigpipe_handler);
		fprintf(stderr,
		    "Connection closed by client while sending document\n");
	    } else {
		perror("send");
	    }
	    return -1;
	}
	buffer += written;
	count -= written;
    } while (count);

    return 0;
}


static int send_doc(int sock, int docsize)
{
    char contlenbuf[32];

    assert(docsize > 0);

    snprintf(contlenbuf, 32, "Content-Length: %d\n\n",
	     docsize * 1024 + ok_leadin_len + ok_leadout_len);

    if (send_all(sock, ok_header, ok_header_len) < 0) {
	return -1;
    }
    if (send_all(sock, contlenbuf, strlen(contlenbuf)) < 0) {
	return -1;
    }
    if (send_all(sock, ok_leadin, ok_leadin_len) < 0) {
	return -1;
    }

    while (docsize) {
	unsigned int bufsiz = docsize > DOCBUFSIZE ? DOCBUFSIZE : docsize;
	if (send_all(sock, docbuf, bufsiz * 1024) < 0) {
	    return -1;
	}

	assert(bufsiz <= docsize);
	docsize -= bufsiz;
    }

    if (send_all(sock, ok_leadout, ok_leadout_len) < 0) {
	return -1;
    }

    return 0;
}


static void usage(const char *me)
{
    fprintf(stderr, "Usage: %s [ OPTIONS ]\n", me);
    fprintf(stderr,
	    "-p NUM    --port=NUM         accept requests on port NUM (default %d)\n",
	    DEFAULT_PORT);
    fprintf(stderr,
	    "-l FILE   --logfile=FILE     log requests to file FILE (default %s)\n",
	    DEFAULT_LOGFILE);
    fprintf(stderr,
	    "-f NUM    --fork=NUM         fork NUM server processes (default 0)\n");
    fprintf(stderr, "-h        --help             this screen\n");
    exit(1);
}


int main(int argc, char *argv[])
{
    struct sockaddr_in me;
    int sock, msgsock, flag;
    unsigned short port = DEFAULT_PORT;
    char *logfile = DEFAULT_LOGFILE;
    int fork_n = 0, n_chld = 0;
    pid_t *children = NULL;
    int c, longind;
    struct option longopts[] = {
	{"port", 1, NULL, 'p'},
	{"logfile", 1, NULL, 'l'},
	{"fork", 1, NULL, 'f'},
	{"help", 0, NULL, 'h'},
	{NULL, 0, NULL, 0}
    };

    while ((c = getopt_long(argc, argv, "p:l:f:h", longopts, &longind)) != EOF) {
	switch (c) {
	case 'p':
	    port = atoi(optarg);
	    break;
	case 'l':
	    logfile = optarg;
	    break;
	case 'f':
	    fork_n = atoi(optarg);
	    break;
	case 'h':
	case '?':
	    usage(argv[0]);
	    break;
	}
    }

    if (init_buffer() < 0) {
	fprintf(stderr, "%s: memory exhausted\n", argv[0]);
	exit(1);
    }

    if (fork_n) {
	children = (pid_t *) calloc(fork_n, sizeof(pid_t));
	if (!children) {
	    fprintf(stderr, "%s: memory exhausted\n", argv[0]);
	    exit(1);
	}
    }

    logfp = fopen(logfile, "a");
    if (!logfp) {
	perror(logfile);
	exit(1);
    }

    signal(SIGPIPE, sigpipe_handler);
    signal(SIGINT, sigterm_handler);
    signal(SIGTERM, sigterm_handler);

    siginterrupt(SIGPIPE, 1);
    siginterrupt(SIGINT, 1);
    siginterrupt(SIGTERM, 1);

    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
	perror("opening stream socket");
	fprintf(logfp, "Error opening socket -- aborting\n");
	fclose(logfp);
	exit(1);
    }

    flag = 1;
    if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (const void *) &flag,
	    sizeof(int)) < 0) {
	perror("setsockopt(SO_REUSEADDR)");
    }

    me.sin_family = AF_INET;
    me.sin_addr.s_addr = INADDR_ANY;
    me.sin_port = htons(port);
    if (bind(sock, (struct sockaddr *) &me, sizeof me) < 0) {
	perror("binding stream socket");
	fprintf(logfp, "Error binding socket -- aborting\n");
	fclose(logfp);
	exit(1);
    }

    ok_header_len = strlen(ok_header);
    ok_leadin_len = strlen(ok_leadin);
    ok_leadout_len = strlen(ok_leadout);

    /* start accepting connections */
    listen(sock, 5);

    printf("*** %s listening on port %u ***\n", VERSION, port);

    if (fork_n) {
	for (n_chld = 0; n_chld < fork_n; n_chld++) {
	    children[n_chld] = fork();
	    if (children[n_chld] < 0) {
		perror("fork");
		break;
	    } else if (children[n_chld] == 0) {
		free(children);
		children = NULL;
		logmsg("Server process starting");
		break;
	    }
	}
    }

    if (children) {
	char logbuf[128];
	int i;

	printf("Forked %d server processes\n", n_chld);
	sprintf(logbuf, "Forked %d server processes", n_chld);
	logmsg(logbuf);

	while (!sigterm) {
	    sleep(60);
	}

	for (i = 0; i < n_chld; i++) {
	    if (waitpid(children[i], NULL, 0) < 0) {
		char logbuf[128];
		perror("waitpid");
		sprintf(logbuf, "waitpid() failed for child process %d",
			children[i]);
		logmsg(logbuf);
	    }
	}
	logmsg("All child processes terminated, exiting");
    } else {
	while (!sigterm) {
	    int docsize;

	    msgsock = accept(sock, NULL, NULL);

	    if (msgsock == -1) {
		if (sigterm) {
		    break;
		}
		/* perror("accept"); */
		continue;
	    }

	    docsize = get_request(msgsock);
	    if (docsize >= 0) {
		if (docsize == 0) {
		    send_all(msgsock, doc404, strlen(doc404));
		} else {
		    send_doc(msgsock, docsize);
		}
	    }

	    close(msgsock);
	}
    }

    close(sock);
    fclose(logfp);

    return 0;
}

